/*--------------------*/
// for encoder
/*--------------------*/
// encoder 0: the base encoder, original 500 CPR -> 6000 CPR
#define encoder0PinA 3
#define encoder0PinB 2
volatile int encoder0Count = 0;
volatile float angle0 = 0;
volatile float velocity0 = 0;
volatile float angle0_previous = 0;
volatile float angle0_post = 0;
// encoder 1: the pendulum encoder, original 1250 CPR -> 5000 CPR
#define encoder1PinA 20
#define encoder1PinB 21
volatile int encoder1Count = 0;
volatile float angle1 = 0;
volatile float velocity1 = 0;
volatile float angle1_previous = 0;
volatile float angle1_post = 0;

/*--------------------*/
// for Timer2
/*--------------------*/
volatile int tcnt2 = 131;
volatile int t = 0;

/*--------------------*/
// for system input
/*--------------------*/
// k's are feedback gains
// result from LQR, when Q = diag([1.5 8 0 0]); R = 0.0028;
volatile float error_proportional = 0;
volatile float error_derivative = 0;
volatile float error_integral = 0;
float kp = 50;
float kd = 8;
float ki = 40;
float u = 0;
float motor_input = 0;
int INPUT_MAX = 3200;

void setup()
{
  /*--------------------*/
  // for motor
  /*--------------------*/
  // initialize software serial object with baud rate of 38.4 kbps
  Serial1.begin(38400);
  // the Simple Motor Controller must be running for at least 1 ms
  // before we try to send serial data, so we delay here for 5 ms
  delay(5);
  // if the Simple Motor Controller has automatic baud detection
  // enabled, we first need to send it the byte 0xAA (170 in decimal)
  // so that it can learn the baud rate
  Serial1.print(0xAA, BYTE);
  // next we need to send the Exit Safe Start command, which
  // clears the safe-start violation and lets the motor run
  exitSafeStart();

  /*--------------------*/
  // for encoder
  /*--------------------*/  
  // for encoder 0
  pinMode(encoder0PinA, INPUT); 
  pinMode(encoder0PinB, INPUT); 
  attachInterrupt(0, doEncoder0A, CHANGE);  
  attachInterrupt(1, doEncoder0B, CHANGE);  
  // for encoder 1
  pinMode(encoder1PinA, INPUT); 
  pinMode(encoder1PinB, INPUT); 
  attachInterrupt(2, doEncoder1A, CHANGE);  
  attachInterrupt(3, doEncoder1B, CHANGE);    
  
  /*--------------------*/
  // for Timer2
  /*--------------------*/
  // interrupts every 1 ms
  TIMSK2 &= ~(1<<TOIE2);
  TCCR2A &= ~((1<<WGM21) | (1<<WGM20));
  TCCR2B &= ~(1<<WGM22);
  ASSR &= ~(1<<AS2);
  TIMSK2 &= ~(1<<OCIE2A);
  TCCR2B |= (1<<CS22)  | (1<<CS20);
  TCCR2B &= ~(1<<CS21);
  TCNT2 = tcnt2;
  TIMSK2 |= (1<<TOIE2);

  //Serial.begin (9600); // for debugging
}

void loop()
{
  u = - (kp*error_proportional + kd*error_derivative + ki*error_integral);
  motor_input = 3200 / 9 * u; // because a 9V battery is used to power the motor
  // 3200 : 9V = motor_input : u, thus motor_input = 3200*u/9
  motor_input = (int)motor_input;
  
  if (motor_input >= INPUT_MAX){
    motor_input = INPUT_MAX;
  }
  else if (motor_input <= -INPUT_MAX){
    motor_input = -INPUT_MAX;
  }
  
  /*
  Serial.print(error_proportional);
  Serial.print(" ");
  Serial.print(error_derivative);
  Serial.print(" ");
  Serial.print(error_integral);
  Serial.print(" ");
  Serial.println(motor_input);
  */
  
  setMotorSpeed(motor_input); // full speed is 3200
}
